# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import inspect
from inspect import Parameter
from typing import Any, Callable, Dict, List, Optional, TypeVar, no_type_check

from typing_extensions import Protocol

from .errors import DIError

_T = TypeVar("_T", covariant=True)


class Component(Protocol[_T]):  # pragma: no cover
    """The component protocol.

    Examples:

      >>> class DBComponent:
      ...   is_cacheable = True
      ...   is_singleton = True
      ...
      ...   def can_handle_parameter(self, parameter: Parameter) -> bool:
      ...     return parameter.annotation is DB
      ...
      ...   def resolve(self, settings: Settings) -> DB:
      ...     return DB(settings["database_dsn"])

    """

    @property
    def is_cacheable(self) -> bool:
        """If True, then the component will be cached within a
        resolver meaning that instances of the resolved component will
        be reused within a single request-response cycle.  This should
        be True for most components.  Defaults to True.
        """

    @property
    def is_singleton(self) -> bool:
        """If True, then the component will be treated as a singleton
        and cached forever after its first use.  Defaults to False.
        """

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        """Returns True when parameter represents the desired component.
        """

    @no_type_check
    def resolve(self) -> _T:
        """Returns an instance of the component.
        """


class DependencyInjector:
    """The dependency injector maintains component state and
    instantiates the resolver.

    Parameters:
      components: The list of components that are used to resolve
        functions' dependencies.
    """

    __slots__ = [
        "components",
        "singletons",
    ]

    components: List[Component[Any]]
    singletons: Dict[Component[Any], Any]

    def __init__(self, components: List[Component[Any]]) -> None:
        self.components = components or []
        self.singletons = {}

        for component in components:
            if getattr(component, "is_singleton", False) and component not in self.singletons:
                resolver = self.get_resolver()
                resolved_component = resolver.resolve(component.resolve)
                self.singletons[component] = resolved_component()

    def get_resolver(self, instances: Optional[Dict[Any, Any]] = None) -> "DependencyResolver":
        """Get the resolver for this Injector.
        """
        return DependencyResolver(
            self.components,
            {**self.singletons, **(instances or {})},
        )


class DependencyResolver:
    """The resolver does the work of actually filling in all of a
    function's dependencies.
    """

    __slots__ = [
        "components",
        "instances",
    ]

    def __init__(self, components: List[Component[Any]], instances: Dict[Component[Any], Any]) -> None:
        self.components = components[:]
        self.instances = instances

    def add_component(self, component: Component[Any]) -> None:
        """Add a component to this resolver without adding it to the
        base dependency injector.  This is useful for runtime-built
        components like RouteParamsComponent.
        """
        self.components.append(component)

    def resolve(
            self,
            fn: Callable[..., Any],
            resolving_parameter: Optional[Parameter] = None,
    ) -> Callable[..., Any]:
        """Resolve a function's dependencies.
        """

        @functools.wraps(fn)
        def resolved_fn(**params: Any) -> Any:
            signature = inspect.signature(fn)
            for parameter in signature.parameters.values():
                if parameter.name in params:
                    continue

                # When Parameter is requested then we assume that the
                # caller actually wants the parameter that resolved the
                # current component that's being resolved.  See QueryParam
                # or Header components for an example.
                if parameter.annotation is Parameter:
                    params[parameter.name] = resolving_parameter
                    continue

                # When a DependencyResolver is requested, then we assume
                # the caller wants this instance.
                if parameter.annotation is DependencyResolver:
                    params[parameter.name] = self
                    continue

                # If our instances contains an exact match for a type,
                # then we return that.  This is used to inject the current
                # Request object among other things.
                try:
                    params[parameter.name] = self.instances[parameter.annotation]
                    continue
                except KeyError:
                    pass

                for component in self.components:
                    if component.can_handle_parameter(parameter):
                        try:
                            params[parameter.name] = self.instances[component]
                        except KeyError:
                            factory = self.resolve(component.resolve, resolving_parameter=parameter)
                            params[parameter.name] = instance = factory()

                            if getattr(component, "is_cacheable", True):
                                self.instances[component] = instance

                        break
                else:
                    raise DIError(f"cannot resolve parameter {parameter} of function {fn}")

            return fn(**params)

        return resolved_fn



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from inspect import Parameter

from molten import HTTP_200, Response

try:
    import jinja2
except ImportError:  # pragma: no cover
    raise ImportError("'jinja2' missing. Run 'pip install jinja2'.")


class Templates:
    """Renders jinja2 templates.
    """

    __slots__ = ["environment"]

    def __init__(self, path: str) -> None:
        self.environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(path),
        )

    def render(self, template_name: str, **context) -> Response:
        """Find a template and render it.

        Parameters:
          template_name: The name of the template to render.
          \**context: Bindings passed to the template.
        """
        template = self.environment.get_template(template_name)
        rendered_template = template.render(**context)
        return Response(HTTP_200, content=rendered_template, headers={
            "content-type": "text/html",
        })


class TemplatesComponent:
    """A component that builds a jinja2 template renderer.

    Parameters:
      path: The path to a folder containing your templates.
    """

    __slots__ = ["path"]

    is_cacheable = True
    is_singleton = True

    def __init__(self, path: str) -> None:
        self.path = path

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Templates

    def resolve(self) -> Templates:
        return Templates(self.path)



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import os
from typing import BinaryIO, Optional, Union

from .cookies import Cookie
from .headers import Headers, HeadersDict


class Response:
    """An HTTP response.

    Parameters:
      status: The status line of the response.
      headers: Optional response headers.
      content: Optional response content as a string.
      stream: Optional response content as a file-like object.
      encoding: An optional encoding for the response.
    """

    __slots__ = [
        "status",
        "headers",
        "stream",
    ]

    def __init__(
            self,
            status: str,
            headers: Optional[Union[HeadersDict, Headers]] = None,
            content: Optional[str] = None,
            stream: Optional[BinaryIO] = None,
            encoding: str = "utf-8",
    ) -> None:
        self.status = status

        if isinstance(headers, dict):
            self.headers = Headers(headers)
        else:
            self.headers = headers or Headers()

        if content is not None:
            self.stream: BinaryIO = io.BytesIO(content.encode(encoding))
        elif stream is not None:
            self.stream: BinaryIO = stream
        else:
            self.stream: BinaryIO = io.BytesIO()

    def get_content_length(self) -> Optional[int]:
        """Compute the content length of this response.
        """
        content_length = self.headers.get_int("content_length")
        if content_length is None:
            try:
                stream_stat = os.fstat(self.stream.fileno())
                content_length = stream_stat.st_size
            except OSError:
                old_position = self.stream.tell()

                try:
                    self.stream.seek(0, os.SEEK_END)
                    content_length = self.stream.tell()
                finally:
                    self.stream.seek(old_position, os.SEEK_SET)

        return content_length

    def set_cookie(self, cookie: Cookie) -> None:
        """Add a cookie to this response.
        """
        self.headers.add("set-cookie", cookie.encode())

    def __repr__(self) -> str:
        return f"Response(status={repr(self.status)}, headers={repr(self.headers)})"



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict, List, Optional, Union
from urllib.parse import parse_qsl

from ..common import MultiDict
from ..errors import ParamMissing
from ..typing import Environ

ParamsDict = Dict[str, Union[str, List[str]]]


class QueryParams(MultiDict[str, str]):
    """A mapping from param names to lists of values.  Once
    constructed, these instances cannot be modified.
    """

    @classmethod
    def from_environ(cls, environ: Environ) -> "QueryParams":
        """Construct a QueryParams instance from a WSGI environ.
        """
        return cls.parse(environ.get("QUERY_STRING", ""))

    @classmethod
    def parse(cls, query_string: str) -> "QueryParams":
        """Construct a QueryParams instance from a query string.
        """
        return cls(parse_qsl(query_string))

    def get(self, name: str, default: Optional[str] = None) -> Optional[str]:
        """Get the last value for a given key.
        """
        try:
            return self[name]
        except ParamMissing:
            return default

    def __getitem__(self, name: str) -> str:
        """Get the last value for a given key.

        Raises:
          ParamMissing: When the key is missing.
        """
        try:
            return self._data[name][-1]
        except IndexError:
            raise ParamMissing(name)



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
from typing import Any

from typing_extensions import Protocol

from .http import Response
from .validation import dump_schema, is_schema


class ResponseRenderer(Protocol):  # pragma: no cover
    """Protocol for response renderers.
    """

    def can_render_response(self, accept: str) -> bool:
        """Returns True if this renderer can render data for the given mime type.
        """

    def render(self, status: str, response_data: Any) -> Response:
        """Attempt to render the response data.
        """


class JSONRenderer:
    """A JSON response renderer.
    """

    def can_render_response(self, accept: str) -> bool:
        return accept.startswith("application/json")

    def render(self, status: str, response_data: Any) -> Response:
        content = json.dumps(response_data, default=self.default)
        return Response(status, content=content, headers={
            "content-type": "application/json; charset=utf-8",
        })

    def default(self, ob: Any) -> Any:
        """You may override this when subclassing the JSON renderer in
        order to encode non-standard object types.
        """
        if is_schema(type(ob)):
            return dump_schema(ob)

        raise TypeError(f"cannot encode values of type {type(ob)}")  # pragma: no cover



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['molten', 'molten.http', 'molten.testing', 'molten.validation']

package_data = \
{'': ['*'], 'molten': ['contrib/*']}

install_requires = \
['typing-extensions>=3.6,<4.0']

setup_kwargs = {
    'name': 'molten',
    'version': '0.0.2',
    'description': 'An experimental web framework for Python 3.6+.',
    'long_description': None,
    'author': 'Bogdan Popa',
    'author_email': 'bogdan@cleartype.io',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Callable, List

from .errors import HeaderMissing, HTTPError, ParseError
from .http import HTTP_200, HTTP_400, HTTP_406, Request, Response
from .renderers import ResponseRenderer


class ResponseRendererMiddleware:
    """A middleware that renders responses.

    Parameters:
      renderers: The list of response renderers that is used to render
        responses.  The first renderer whose can_render_response
        method returns True is used to render the current response.
    """

    def __init__(self, renderers: List[ResponseRenderer]) -> None:
        self.renderers = renderers

    def __call__(self, handler: Callable[..., Any]) -> Callable[..., Response]:
        def handle(request: Request) -> Response:
            try:
                response = handler()
                if isinstance(response, Response):
                    return response

                if isinstance(response, tuple):
                    status, response = response

                else:
                    status, response = HTTP_200, response

            except ParseError as e:
                status, response = HTTP_400, {"error": str(e)}

            except HTTPError as e:
                status, response = e.status, e.response

            try:
                accept = request.headers["accept"]
            except HeaderMissing:
                accept = "*/*"

            for renderer in self.renderers:
                if accept == "*/*" or renderer.can_render_response(accept):
                    return renderer.render(status, response)

            return Response(HTTP_406, content="Not Acceptable", headers={
                "content-type": "text/plain",
            })

        return handle
